"""
module catan provides classes and enums useful for representing a catan game.

The main class is Game, which contains Players, a Board, and a CatanLog.

See module boardbuilder for the mechanics of creating and modifying Board objects.

All classes in this module:
- Game
- Player
- Board
- Tile
- Terrain
- HexNumber
- Port
- Piece
- PieceType
"""
